@@ -1,5 +1,5 @@
 /* libFLAC - Free Lossless Audio Codec library
- * Copyright (C) 2000,2001  Josh Coalson
+ * Copyright (C) 2000,2001,2002  Josh Coalson
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public
@@ -23,11 +23,10 @@
 #include "assert.h"
 #include "file_decoder.h"
 #include "format.h"
+#include "metadata.h"
 #include "ordinals.h"
-#include "seek_table.h"
 #include "seekable_stream_decoder.h"
 #include "stream_decoder.h"
 #include "stream_encoder.h"
-#include "utility.h"
 
 #endif
