@@ -1,5 +1,5 @@
 /* flac - Command-line FLAC encoder/decoder
- * Copyright (C) 2000,2001  Josh Coalson
+ * Copyright (C) 2000,2001,2002  Josh Coalson
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -86,17 +86,17 @@ typedef struct {
 	FLAC__uint64 stream_offset; /* i.e. number of bytes before the first byte of the the first frame's header */
 	unsigned current_frame;
 	verify_fifo_struct verify_fifo;
-	FLAC__StreamMetaData_SeekTable seek_table;
+	FLAC__StreamMetadata_SeekTable seek_table;
 	unsigned first_seek_point_to_check;
-	FLAC__bool use_ogg;
 #ifdef FLAC__HAS_OGG
+	FLAC__bool use_ogg;
 	ogg_info_struct ogg;
 #endif
 } encoder_wrapper_struct;
 
 static FLAC__bool is_big_endian_host;
 
-static unsigned char ucbuffer[CHUNK_OF_SAMPLES*FLAC__MAX_CHANNELS*((FLAC__MAX_BITS_PER_SAMPLE+7)/8)];
+static unsigned char ucbuffer[CHUNK_OF_SAMPLES*FLAC__MAX_CHANNELS*((FLAC__REFERENCE_CODEC_MAX_BITS_PER_SAMPLE+7)/8)];
 static signed char *scbuffer = (signed char *)ucbuffer;
 static FLAC__uint16 *usbuffer = (FLAC__uint16 *)ucbuffer;
 static FLAC__int16 *ssbuffer = (FLAC__int16 *)ucbuffer;
@@ -107,16 +107,16 @@ static FLAC__int32 *input[FLAC__MAX_CHANNELS];
 /* local routines */
 static FLAC__bool init(encoder_wrapper_struct *encoder_wrapper);
 static FLAC__bool init_encoder(encode_options_t options, unsigned channels, unsigned bps, unsigned sample_rate, encoder_wrapper_struct *encoder_wrapper);
-static FLAC__bool convert_to_seek_table(char *requested_seek_points, int num_requested_seek_points, FLAC__uint64 stream_samples, unsigned blocksize, FLAC__StreamMetaData_SeekTable *seek_table);
-static void append_point_to_seek_table(FLAC__StreamMetaData_SeekTable *seek_table, FLAC__uint64 sample, FLAC__uint64 stream_samples, FLAC__uint64 blocksize);
-static int seekpoint_compare(const FLAC__StreamMetaData_SeekPoint *l, const FLAC__StreamMetaData_SeekPoint *r);
+static FLAC__bool convert_to_seek_table(char *requested_seek_points, int num_requested_seek_points, FLAC__uint64 stream_samples, unsigned blocksize, FLAC__StreamMetadata_SeekTable *seek_table);
+static void append_point_to_seek_table(FLAC__StreamMetadata_SeekTable *seek_table, FLAC__uint64 sample, FLAC__uint64 stream_samples, FLAC__uint64 blocksize);
+static int seekpoint_compare(const FLAC__StreamMetadata_SeekPoint *l, const FLAC__StreamMetadata_SeekPoint *r);
 static void format_input(FLAC__int32 *dest[], unsigned wide_samples, FLAC__bool is_big_endian, FLAC__bool is_unsigned_samples, unsigned channels, unsigned bps, encoder_wrapper_struct *encoder_wrapper);
-static void append_to_verify_fifo(encoder_wrapper_struct *encoder_wrapper, const FLAC__int32 *input[], unsigned channels, unsigned wide_samples);
+static void append_to_verify_fifo(encoder_wrapper_struct *encoder_wrapper, const FLAC__int32 * const input[], unsigned channels, unsigned wide_samples);
 static FLAC__StreamEncoderWriteStatus write_callback(const FLAC__StreamEncoder *encoder, const FLAC__byte buffer[], unsigned bytes, unsigned samples, unsigned current_frame, void *client_data);
-static void metadata_callback(const FLAC__StreamEncoder *encoder, const FLAC__StreamMetaData *metadata, void *client_data);
+static void metadata_callback(const FLAC__StreamEncoder *encoder, const FLAC__StreamMetadata *metadata, void *client_data);
 static FLAC__StreamDecoderReadStatus verify_read_callback(const FLAC__StreamDecoder *decoder, FLAC__byte buffer[], unsigned *bytes, void *client_data);
-static FLAC__StreamDecoderWriteStatus verify_write_callback(const FLAC__StreamDecoder *decoder, const FLAC__Frame *frame, const FLAC__int32 *buffer[], void *client_data);
-static void verify_metadata_callback(const FLAC__StreamDecoder *decoder, const FLAC__StreamMetaData *metadata, void *client_data);
+static FLAC__StreamDecoderWriteStatus verify_write_callback(const FLAC__StreamDecoder *decoder, const FLAC__Frame *frame, const FLAC__int32 * const buffer[], void *client_data);
+static void verify_metadata_callback(const FLAC__StreamDecoder *decoder, const FLAC__StreamMetadata *metadata, void *client_data);
 static void verify_error_callback(const FLAC__StreamDecoder *decoder, FLAC__StreamDecoderErrorStatus status, void *client_data);
 static void print_stats(const encoder_wrapper_struct *encoder_wrapper);
 static FLAC__bool read_little_endian_uint16(FILE *f, FLAC__uint16 *val, FLAC__bool eof_ok, const char *fn);
@@ -156,12 +156,13 @@ int flac__encode_wav(FILE *infile, long infilesize, const char *infilename, cons
 	(void)lookahead_length;
 
 	if(0 == strcmp(outfilename, "-")) {
-		encoder_wrapper.fout = stdout;
+		encoder_wrapper.fout = file__get_binary_stdout();
 	}
 	else {
 		if(0 == (encoder_wrapper.fout = fopen(outfilename, "wb"))) {
 			fprintf(stderr, "%s: ERROR: can't open output file %s\n", encoder_wrapper.inbasefilename, outfilename);
-			fclose(infile);
+			if(infile != stdin)
+				fclose(infile);
 			return 1;
 		}
 	}
@@ -215,7 +216,7 @@ int flac__encode_wav(FILE *infile, long infilesize, const char *infilename, cons
 				/* sample rate */
 				if(!read_little_endian_uint32(infile, &xx, false, encoder_wrapper.inbasefilename))
 					goto wav_abort_;
-				if(xx == 0 || xx > FLAC__MAX_SAMPLE_RATE) {
+				if(!FLAC__format_is_valid_sample_rate(xx)) {
 					fprintf(stderr, "%s: ERROR: unsupported sample rate %u\n", encoder_wrapper.inbasefilename, (unsigned)xx);
 					goto wav_abort_;
 				}
@@ -233,7 +234,7 @@ int flac__encode_wav(FILE *infile, long infilesize, const char *infilename, cons
 				/* bits per sample */
 				if(!read_little_endian_uint16(infile, &x, false, encoder_wrapper.inbasefilename))
 					goto wav_abort_;
-				if(x != 8 && x != 16) {
+				if(x != 8 && x != 16 && x != 24) {
 					fprintf(stderr, "%s: ERROR: unsupported bits per sample %u\n", encoder_wrapper.inbasefilename, (unsigned)x);
 					goto wav_abort_;
 				}
@@ -262,7 +263,7 @@ int flac__encode_wav(FILE *infile, long infilesize, const char *infilename, cons
 				fprintf(stderr, "%s: WARNING: skipping extra 'data' sub-chunk\n", encoder_wrapper.inbasefilename);
 			}
 			else if(!got_fmt_chunk) {
-				fprintf(stderr, "%s: ERROR: got data sub-chunk before fmt sub-chunk\n", encoder_wrapper.inbasefilename);
+				fprintf(stderr, "%s: ERROR: got 'data' sub-chunk before 'fmt' sub-chunk\n", encoder_wrapper.inbasefilename);
 				goto wav_abort_;
 			}
 			else {
@@ -274,13 +275,8 @@ int flac__encode_wav(FILE *infile, long infilesize, const char *infilename, cons
 				bytes_per_wide_sample = channels * (bps >> 3);
 
 				if(options.common.skip > 0) {
-					if(infile != stdin) {
-						if(-1 == fseek(infile, bytes_per_wide_sample * (unsigned)options.common.skip, SEEK_CUR)) {
-							fprintf(stderr, "%s: ERROR during seek while skipping samples\n", encoder_wrapper.inbasefilename);
-							goto wav_abort_;
-						}
-					}
-					else {
+					if(fseek(infile, bytes_per_wide_sample * (unsigned)options.common.skip, SEEK_CUR) < 0) {
+						/* can't seek input, read ahead manually... */
 						unsigned left, need;
 						for(left = (unsigned)options.common.skip; left > 0; ) { /*@@@ WATCHOUT: 4GB limit */
 							need = min(left, CHUNK_OF_SAMPLES);
@@ -315,11 +311,9 @@ int flac__encode_wav(FILE *infile, long infilesize, const char *infilename, cons
 				 * first do any samples in the reservoir
 				 */
 				if(options.sector_align && *options.align_reservoir_samples > 0) {
-					/* NOTE: some versions of GCC can't figure out const-ness right and will give you an 'incompatible pointer type' warning on arg 2 here: */
-					append_to_verify_fifo(&encoder_wrapper, options.align_reservoir, channels, *options.align_reservoir_samples);
+					append_to_verify_fifo(&encoder_wrapper, (const FLAC__int32 * const *)options.align_reservoir, channels, *options.align_reservoir_samples);
 
-					/* NOTE: some versions of GCC can't figure out const-ness right and will give you an 'incompatible pointer type' warning on arg 2 here: */
-					if(!FLAC__stream_encoder_process(encoder_wrapper.encoder, options.align_reservoir, *options.align_reservoir_samples)) {
+					if(!FLAC__stream_encoder_process(encoder_wrapper.encoder, (const FLAC__int32 * const *)options.align_reservoir, *options.align_reservoir_samples)) {
 						fprintf(stderr, "%s: ERROR during encoding, state = %d:%s\n", encoder_wrapper.inbasefilename, FLAC__stream_encoder_get_state(encoder_wrapper.encoder), FLAC__StreamEncoderStateString[FLAC__stream_encoder_get_state(encoder_wrapper.encoder)]);
 						goto wav_abort_;
 					}
@@ -362,8 +356,7 @@ int flac__encode_wav(FILE *infile, long infilesize, const char *infilename, cons
 							unsigned wide_samples = bytes_read / bytes_per_wide_sample;
 							format_input(input, wide_samples, false, is_unsigned_samples, channels, bps, &encoder_wrapper);
 
-							/* NOTE: some versions of GCC can't figure out const-ness right and will give you an 'incompatible pointer type' warning on arg 2 here: */
-							if(!FLAC__stream_encoder_process(encoder_wrapper.encoder, input, wide_samples)) {
+							if(!FLAC__stream_encoder_process(encoder_wrapper.encoder, (const FLAC__int32 * const *)input, wide_samples)) {
 								fprintf(stderr, "%s: ERROR during encoding, state = %d:%s\n", encoder_wrapper.inbasefilename, FLAC__stream_encoder_get_state(encoder_wrapper.encoder), FLAC__StreamEncoderStateString[FLAC__stream_encoder_get_state(encoder_wrapper.encoder)]);
 								goto wav_abort_;
 							}
@@ -385,11 +378,9 @@ int flac__encode_wav(FILE *infile, long infilesize, const char *infilename, cons
 							data_bytes = wide_samples * bytes_per_wide_sample;
 							for(channel = 0; channel < channels; channel++)
 								memset(input[channel], 0, data_bytes);
-							/* NOTE: some versions of GCC can't figure out const-ness right and will give you an 'incompatible pointer type' warning on arg 2 here: */
-							append_to_verify_fifo(&encoder_wrapper, input, channels, wide_samples);
+							append_to_verify_fifo(&encoder_wrapper, (const FLAC__int32 * const *)input, channels, wide_samples);
 
-							/* NOTE: some versions of GCC can't figure out const-ness right and will give you an 'incompatible pointer type' warning on arg 2 here: */
-							if(!FLAC__stream_encoder_process(encoder_wrapper.encoder, input, wide_samples)) {
+							if(!FLAC__stream_encoder_process(encoder_wrapper.encoder, (const FLAC__int32 * const *)input, wide_samples)) {
 								fprintf(stderr, "%s: ERROR during encoding, state = %d:%s\n", encoder_wrapper.inbasefilename, FLAC__stream_encoder_get_state(encoder_wrapper.encoder), FLAC__StreamEncoderStateString[FLAC__stream_encoder_get_state(encoder_wrapper.encoder)]);
 								goto wav_abort_;
 							}
@@ -404,7 +395,7 @@ int flac__encode_wav(FILE *infile, long infilesize, const char *infilename, cons
 								goto wav_abort_;
 							}
 							else if(bytes_read != (*options.align_reservoir_samples) * bytes_per_wide_sample) {
-								fprintf(stderr, "%s: WARNING: unexpected EOF; expected %u samples, got %u samples\n", encoder_wrapper.inbasefilename, (unsigned)encoder_wrapper.total_samples_to_encode, (unsigned)encoder_wrapper.samples_written);
+								fprintf(stderr, "%s: WARNING: unexpected EOF; read %u bytes; expected %u samples, got %u samples\n", encoder_wrapper.inbasefilename, (unsigned)bytes_read, (unsigned)encoder_wrapper.total_samples_to_encode, (unsigned)encoder_wrapper.samples_written);
 								data_bytes = 0;
 							}
 							else {
@@ -423,13 +414,8 @@ int flac__encode_wav(FILE *infile, long infilesize, const char *infilename, cons
 			/* sub-chunk size */
 			if(!read_little_endian_uint32(infile, &xx, false, encoder_wrapper.inbasefilename))
 				goto wav_abort_;
-			if(infile != stdin) {
-				if(-1 == fseek(infile, xx, SEEK_CUR)) {
-					fprintf(stderr, "%s: ERROR during seek while skipping unsupported sub-chunk\n", encoder_wrapper.inbasefilename);
-					goto wav_abort_;
-				}
-			}
-			else {
+			if(fseek(infile, xx, SEEK_CUR) < 0) {
+				/* can't seek input, read ahead manually... */
 				unsigned left, need;
 				const unsigned chunk = sizeof(ucbuffer);
 				for(left = xx; left > 0; ) {
@@ -445,8 +431,7 @@ int flac__encode_wav(FILE *infile, long infilesize, const char *infilename, cons
 	}
 
 	if(encoder_wrapper.encoder) {
-		if(FLAC__stream_encoder_get_state(encoder_wrapper.encoder) == FLAC__STREAM_ENCODER_OK)
-			FLAC__stream_encoder_finish(encoder_wrapper.encoder);
+		FLAC__stream_encoder_finish(encoder_wrapper.encoder);
 		FLAC__stream_encoder_delete(encoder_wrapper.encoder);
 #ifdef FLAC__HAS_OGG
 		if(encoder_wrapper.use_ogg)
@@ -478,8 +463,7 @@ wav_abort_:
 	if(encoder_wrapper.verbose && encoder_wrapper.total_samples_to_encode > 0)
 		fprintf(stderr, "\n");
 	if(encoder_wrapper.encoder) {
-		if(FLAC__stream_encoder_get_state(encoder_wrapper.encoder) == FLAC__STREAM_ENCODER_OK)
-			FLAC__stream_encoder_finish(encoder_wrapper.encoder);
+		FLAC__stream_encoder_finish(encoder_wrapper.encoder);
 		FLAC__stream_encoder_delete(encoder_wrapper.encoder);
 #ifdef FLAC__HAS_OGG
 		if(encoder_wrapper.use_ogg)
@@ -523,12 +507,13 @@ int flac__encode_raw(FILE *infile, long infilesize, const char *infilename, cons
 #endif
 
 	if(0 == strcmp(outfilename, "-")) {
-		encoder_wrapper.fout = stdout;
+		encoder_wrapper.fout = file__get_binary_stdout();
 	}
 	else {
 		if(0 == (encoder_wrapper.fout = fopen(outfilename, "wb"))) {
 			fprintf(stderr, "ERROR: can't open output file %s\n", outfilename);
-			fclose(infile);
+			if(infile != stdin)
+				fclose(infile);
 			return 1;
 		}
 	}
@@ -553,13 +538,8 @@ int flac__encode_raw(FILE *infile, long infilesize, const char *infilename, cons
 		if(skip_bytes > lookahead_length) {
 			skip_bytes -= lookahead_length;
 			lookahead_length = 0;
-			if(infile != stdin) {
-				if(-1 == fseek(infile, (long)skip_bytes, SEEK_CUR)) {
-					fprintf(stderr, "%s: ERROR during seek while skipping samples\n", encoder_wrapper.inbasefilename);
-					goto raw_abort_;
-				}
-			}
-			else {
+			if(fseek(infile, (long)skip_bytes, SEEK_CUR) < 0) {
+				/* can't seek input, read ahead manually... */
 				unsigned left, need;
 				const unsigned chunk = sizeof(ucbuffer);
 				for(left = skip_bytes; left > 0; ) {
@@ -611,8 +591,7 @@ int flac__encode_raw(FILE *infile, long infilesize, const char *infilename, cons
 			unsigned wide_samples = bytes_read / bytes_per_wide_sample;
 			format_input(input, wide_samples, options.is_big_endian, options.is_unsigned_samples, options.channels, options.bps, &encoder_wrapper);
 
-			/* NOTE: some versions of GCC can't figure out const-ness right and will give you an 'incompatible pointer type' warning on arg 2 here: */
-			if(!FLAC__stream_encoder_process(encoder_wrapper.encoder, input, wide_samples)) {
+			if(!FLAC__stream_encoder_process(encoder_wrapper.encoder, (const FLAC__int32 * const *)input, wide_samples)) {
 				fprintf(stderr, "%s: ERROR during encoding, state = %d:%s\n", encoder_wrapper.inbasefilename, FLAC__stream_encoder_get_state(encoder_wrapper.encoder), FLAC__StreamEncoderStateString[FLAC__stream_encoder_get_state(encoder_wrapper.encoder)]);
 				goto raw_abort_;
 			}
@@ -620,8 +599,7 @@ int flac__encode_raw(FILE *infile, long infilesize, const char *infilename, cons
 	}
 
 	if(encoder_wrapper.encoder) {
-		if(FLAC__stream_encoder_get_state(encoder_wrapper.encoder) == FLAC__STREAM_ENCODER_OK)
-			FLAC__stream_encoder_finish(encoder_wrapper.encoder);
+		FLAC__stream_encoder_finish(encoder_wrapper.encoder);
 		FLAC__stream_encoder_delete(encoder_wrapper.encoder);
 #ifdef FLAC__HAS_OGG
 		if(encoder_wrapper.use_ogg)
@@ -649,8 +627,7 @@ raw_abort_:
 	if(encoder_wrapper.verbose && encoder_wrapper.total_samples_to_encode > 0)
 		fprintf(stderr, "\n");
 	if(encoder_wrapper.encoder) {
-		if(FLAC__stream_encoder_get_state(encoder_wrapper.encoder) == FLAC__STREAM_ENCODER_OK)
-			FLAC__stream_encoder_finish(encoder_wrapper.encoder);
+		FLAC__stream_encoder_finish(encoder_wrapper.encoder);
 		FLAC__stream_encoder_delete(encoder_wrapper.encoder);
 #ifdef FLAC__HAS_OGG
 		if(encoder_wrapper.use_ogg)
@@ -704,7 +681,9 @@ FLAC__bool init(encoder_wrapper_struct *encoder_wrapper)
 
 FLAC__bool init_encoder(encode_options_t options, unsigned channels, unsigned bps, unsigned sample_rate, encoder_wrapper_struct *encoder_wrapper)
 {
-	unsigned i;
+	unsigned i, num_metadata;
+	FLAC__StreamMetadata seek_table, padding;
+	FLAC__StreamMetadata *metadata[2];
 
 	if(channels != 2)
 		options.do_mid_side = options.loose_mid_side = false;
@@ -744,6 +723,21 @@ FLAC__bool init_encoder(encode_options_t options, unsigned channels, unsigned bp
 		return false;
 	}
 
+	num_metadata = 0;
+	if(encoder_wrapper->seek_table.num_points > 0) {
+		seek_table.is_last = false; /* the encoder will set this for us */
+		seek_table.type = FLAC__METADATA_TYPE_SEEKTABLE;
+		seek_table.length = encoder_wrapper->seek_table.num_points * FLAC__STREAM_METADATA_SEEKPOINT_LENGTH;
+		seek_table.data.seek_table = encoder_wrapper->seek_table;
+		metadata[num_metadata++] = &seek_table;
+	}
+	if(options.padding > 0) {
+		padding.is_last = false; /* the encoder will set this for us */
+		padding.type = FLAC__METADATA_TYPE_PADDING;
+		padding.length = (unsigned)options.padding;
+		metadata[num_metadata++] = &padding;
+	}
+
 	FLAC__stream_encoder_set_streamable_subset(encoder_wrapper->encoder, !options.lax);
 	FLAC__stream_encoder_set_do_mid_side_stereo(encoder_wrapper->encoder, options.do_mid_side);
 	FLAC__stream_encoder_set_loose_mid_side_stereo(encoder_wrapper->encoder, options.loose_mid_side);
@@ -760,9 +754,7 @@ FLAC__bool init_encoder(encode_options_t options, unsigned channels, unsigned bp
 	FLAC__stream_encoder_set_max_residual_partition_order(encoder_wrapper->encoder, options.max_residual_partition_order);
 	FLAC__stream_encoder_set_rice_parameter_search_dist(encoder_wrapper->encoder, options.rice_parameter_search_dist);
 	FLAC__stream_encoder_set_total_samples_estimate(encoder_wrapper->encoder, encoder_wrapper->total_samples_to_encode);
-	FLAC__stream_encoder_set_seek_table(encoder_wrapper->encoder, (encoder_wrapper->seek_table.num_points > 0)? &encoder_wrapper->seek_table : 0);
-	FLAC__stream_encoder_set_padding(encoder_wrapper->encoder, options.padding);
-	FLAC__stream_encoder_set_last_metadata_is_last(encoder_wrapper->encoder, true);
+	FLAC__stream_encoder_set_metadata(encoder_wrapper->encoder, (num_metadata > 0)? metadata : 0, num_metadata);
 	FLAC__stream_encoder_set_write_callback(encoder_wrapper->encoder, write_callback);
 	FLAC__stream_encoder_set_metadata_callback(encoder_wrapper->encoder, metadata_callback);
 	FLAC__stream_encoder_set_client_data(encoder_wrapper->encoder, encoder_wrapper);
@@ -778,7 +770,7 @@ FLAC__bool init_encoder(encode_options_t options, unsigned channels, unsigned bp
 	return true;
 }
 
-FLAC__bool convert_to_seek_table(char *requested_seek_points, int num_requested_seek_points, FLAC__uint64 stream_samples, unsigned blocksize, FLAC__StreamMetaData_SeekTable *seek_table)
+FLAC__bool convert_to_seek_table(char *requested_seek_points, int num_requested_seek_points, FLAC__uint64 stream_samples, unsigned blocksize, FLAC__StreamMetadata_SeekTable *seek_table)
 {
 	unsigned i, j, real_points, placeholders;
 	char *pt = requested_seek_points, *q;
@@ -818,7 +810,7 @@ FLAC__bool convert_to_seek_table(char *requested_seek_points, int num_requested_
 	pt = requested_seek_points;
 
 	/* make some space */
-	if(0 == (seek_table->points = (FLAC__StreamMetaData_SeekPoint*)malloc(sizeof(FLAC__StreamMetaData_SeekPoint) * (real_points+placeholders))))
+	if(0 == (seek_table->points = (FLAC__StreamMetadata_SeekPoint*)malloc(sizeof(FLAC__StreamMetadata_SeekPoint) * (real_points+placeholders))))
 		return false;
 
 	/* initialize the seek_table.  we set frame_samples to zero to signify the points have not yet been hit by a frame write yet. */
@@ -852,7 +844,7 @@ FLAC__bool convert_to_seek_table(char *requested_seek_points, int num_requested_
 	}
 
 	/* sort the seekpoints */
-	qsort(seek_table->points, seek_table->num_points, sizeof(FLAC__StreamMetaData_SeekPoint), (int (*)(const void *, const void *))seekpoint_compare);
+	qsort(seek_table->points, seek_table->num_points, sizeof(FLAC__StreamMetadata_SeekPoint), (int (*)(const void *, const void *))seekpoint_compare);
 
 	/* uniqify the seekpoints */
 	first = true;
@@ -874,7 +866,7 @@ FLAC__bool convert_to_seek_table(char *requested_seek_points, int num_requested_
 	return true;
 }
 
-void append_point_to_seek_table(FLAC__StreamMetaData_SeekTable *seek_table, FLAC__uint64 sample, FLAC__uint64 stream_samples, FLAC__uint64 blocksize)
+void append_point_to_seek_table(FLAC__StreamMetadata_SeekTable *seek_table, FLAC__uint64 sample, FLAC__uint64 stream_samples, FLAC__uint64 blocksize)
 {
 	const FLAC__uint64 target_sample = (sample / blocksize) * blocksize;
 
@@ -882,7 +874,7 @@ void append_point_to_seek_table(FLAC__StreamMetaData_SeekTable *seek_table, FLAC
 		seek_table->points[seek_table->num_points++].sample_number = target_sample;
 }
 
-int seekpoint_compare(const FLAC__StreamMetaData_SeekPoint *l, const FLAC__StreamMetaData_SeekPoint *r)
+int seekpoint_compare(const FLAC__StreamMetadata_SeekPoint *l, const FLAC__StreamMetadata_SeekPoint *r)
 {
 	/* we don't just 'return l->sample_number - r->sample_number' since the result (FLAC__int64) might overflow an 'int' */
 	if(l->sample_number == r->sample_number)
@@ -962,16 +954,15 @@ void format_input(FLAC__int32 *dest[], unsigned wide_samples, FLAC__bool is_big_
 		FLAC__ASSERT(0);
 	}
 
-	/* NOTE: some versions of GCC can't figure out const-ness right and will give you an 'incompatible pointer type' warning on arg 2 here: */
-	append_to_verify_fifo(encoder_wrapper, dest, channels, wide_samples);
+	append_to_verify_fifo(encoder_wrapper, (const FLAC__int32 * const *)dest, channels, wide_samples);
 }
 
-void append_to_verify_fifo(encoder_wrapper_struct *encoder_wrapper, const FLAC__int32 *src[], unsigned channels, unsigned wide_samples)
+void append_to_verify_fifo(encoder_wrapper_struct *encoder_wrapper, const FLAC__int32 * const input[], unsigned channels, unsigned wide_samples)
 {
 	if(encoder_wrapper->verify) {
 		unsigned channel;
 		for(channel = 0; channel < channels; channel++)
-			memcpy(&encoder_wrapper->verify_fifo.original[channel][encoder_wrapper->verify_fifo.tail], src[channel], sizeof(FLAC__int32) * wide_samples);
+			memcpy(&encoder_wrapper->verify_fifo.original[channel][encoder_wrapper->verify_fifo.tail], input[channel], sizeof(FLAC__int32) * wide_samples);
 		encoder_wrapper->verify_fifo.tail += wide_samples;
 		FLAC__ASSERT(encoder_wrapper->verify_fifo.tail <= encoder_wrapper->verify_fifo.size);
 	}
@@ -1033,7 +1024,13 @@ FLAC__StreamEncoderWriteStatus write_callback(const FLAC__StreamEncoder *encoder
 
 		memset(&op, 0, sizeof(op));
 		op.packet = (unsigned char *)buffer;
-		op.packetno = encoder_wrapper->current_frame;
+		op.granulepos = encoder_wrapper->samples_written - 1;
+		/*@@@ WATCHOUT:
+		 * this depends on the behavior of libFLAC that we will get one
+		 * write_callback first with all the metadata (and 'samples'
+		 * will be 0), then one write_callback for each frame.
+		 */
+		op.packetno = (samples == 0? -1 : (int)encoder_wrapper->current_frame);
 		op.bytes = bytes;
 
 		if (encoder_wrapper->bytes_written == bytes)
@@ -1067,11 +1064,11 @@ FLAC__StreamEncoderWriteStatus write_callback(const FLAC__StreamEncoder *encoder
 	}
 }
 
-void metadata_callback(const FLAC__StreamEncoder *encoder, const FLAC__StreamMetaData *metadata, void *client_data)
+void metadata_callback(const FLAC__StreamEncoder *encoder, const FLAC__StreamMetadata *metadata, void *client_data)
 {
 	encoder_wrapper_struct *encoder_wrapper = (encoder_wrapper_struct *)client_data;
 	FLAC__byte b;
-	FILE *f;
+	FILE *f = encoder_wrapper->fout;
 	const FLAC__uint64 samples = metadata->data.stream_info.total_samples;
 	const unsigned min_framesize = metadata->data.stream_info.min_framesize;
 	const unsigned max_framesize = metadata->data.stream_info.max_framesize;
@@ -1084,8 +1081,10 @@ void metadata_callback(const FLAC__StreamEncoder *encoder, const FLAC__StreamMet
 	 * update are not necessary with Ogg as the transport.  We can't do
 	 * it reliably anyway without knowing the Ogg structure.
 	 */
+#ifdef FLAC__HAS_OGG
 	if(encoder_wrapper->use_ogg)
 		return;
+#endif
 
 	/*
 	 * we get called by the encoder when the encoding process has
@@ -1095,25 +1094,26 @@ void metadata_callback(const FLAC__StreamEncoder *encoder, const FLAC__StreamMet
 
 	(void)encoder; /* silence compiler warning about unused parameter */
 
-	if(encoder_wrapper->fout == stdout)
-		return;
-
-	fclose(encoder_wrapper->fout);
-	if(0 == (f = fopen(encoder_wrapper->outfilename, "r+b")))
-		return;
+	if(f != stdout) {
+		fclose(encoder_wrapper->fout);
+		if(0 == (f = fopen(encoder_wrapper->outfilename, "r+b")))
+			return;
+	}
 
 	/* all this is based on intimate knowledge of the stream header
 	 * layout, but a change to the header format that would break this
 	 * would also break all streams encoded in the previous format.
 	 */
 
-	if(-1 == fseek(f, 26, SEEK_SET)) goto samples_;
+	if(-1 == fseek(f, 26, SEEK_SET)) goto end_;
 	fwrite(metadata->data.stream_info.md5sum, 1, 16, f);
 
-samples_:
-	if(-1 == fseek(f, 21, SEEK_SET)) goto framesize_;
+	/* if we get this far we know we can seek so no need to check the
+	 * return value from fseek()
+	 */
+	fseek(f, 21, SEEK_SET);
 	if(fread(&b, 1, 1, f) != 1) goto framesize_;
-	if(-1 == fseek(f, 21, SEEK_SET)) goto framesize_;
+	fseek(f, 21, SEEK_SET);
 	b = (b & 0xf0) | (FLAC__byte)((samples >> 32) & 0x0F);
 	if(fwrite(&b, 1, 1, f) != 1) goto framesize_;
 	b = (FLAC__byte)((samples >> 24) & 0xFF);
@@ -1126,7 +1126,7 @@ samples_:
 	if(fwrite(&b, 1, 1, f) != 1) goto framesize_;
 
 framesize_:
-	if(-1 == fseek(f, 12, SEEK_SET)) goto seektable_;
+	fseek(f, 12, SEEK_SET);
 	b = (FLAC__byte)((min_framesize >> 16) & 0xFF);
 	if(fwrite(&b, 1, 1, f) != 1) goto seektable_;
 	b = (FLAC__byte)((min_framesize >> 8) & 0xFF);
@@ -1157,7 +1157,7 @@ seektable_:
 		pos = (FLAC__STREAM_SYNC_LEN + FLAC__STREAM_METADATA_IS_LAST_LEN + FLAC__STREAM_METADATA_TYPE_LEN + FLAC__STREAM_METADATA_LENGTH_LEN) / 8;
 		pos += metadata->length;
 		pos += (FLAC__STREAM_METADATA_IS_LAST_LEN + FLAC__STREAM_METADATA_TYPE_LEN + FLAC__STREAM_METADATA_LENGTH_LEN) / 8;
-		if(-1 == fseek(f, pos, SEEK_SET)) goto end_;
+		fseek(f, pos, SEEK_SET);
 		for(i = 0; i < encoder_wrapper->seek_table.num_points; i++) {
 			if(!write_big_endian_uint64(f, encoder_wrapper->seek_table.points[i].sample_number)) goto end_;
 			if(!write_big_endian_uint64(f, encoder_wrapper->seek_table.points[i].stream_offset)) goto end_;
@@ -1186,10 +1186,10 @@ FLAC__StreamDecoderReadStatus verify_read_callback(const FLAC__StreamDecoder *de
 		encoder_wrapper->verify_fifo.encoded_bytes -= *bytes;
 	}
 
-	return FLAC__STREAM_DECODER_READ_CONTINUE;
+	return FLAC__STREAM_DECODER_READ_STATUS_CONTINUE;
 }
 
-FLAC__StreamDecoderWriteStatus verify_write_callback(const FLAC__StreamDecoder *decoder, const FLAC__Frame *frame, const FLAC__int32 *buffer[], void *client_data)
+FLAC__StreamDecoderWriteStatus verify_write_callback(const FLAC__StreamDecoder *decoder, const FLAC__Frame *frame, const FLAC__int32 * const buffer[], void *client_data)
 {
 	encoder_wrapper_struct *encoder_wrapper = (encoder_wrapper_struct *)client_data;
 	unsigned channel, l, r;
@@ -1216,7 +1216,7 @@ FLAC__StreamDecoderWriteStatus verify_write_callback(const FLAC__StreamDecoder *
 			FLAC__ASSERT(l < r);
 			FLAC__ASSERT(frame->header.number_type == FLAC__FRAME_NUMBER_TYPE_SAMPLE_NUMBER);
 			fprintf(stderr, "       Absolute sample=%u, frame=%u, channel=%u, sample=%u, expected %d, got %d\n", (unsigned)frame->header.number.sample_number + sample, (unsigned)frame->header.number.sample_number / FLAC__stream_decoder_get_blocksize(decoder), channel, sample, expect, got); /*@@@ WATCHOUT: 4GB limit */
-			return FLAC__STREAM_DECODER_WRITE_ABORT;
+			return FLAC__STREAM_DECODER_WRITE_STATUS_ABORT;
 		}
 	}
 	/* dequeue the frame from the fifo */
@@ -1226,10 +1226,10 @@ FLAC__StreamDecoderWriteStatus verify_write_callback(const FLAC__StreamDecoder *
 		}
 	}
 	encoder_wrapper->verify_fifo.tail -= frame->header.blocksize;
-	return FLAC__STREAM_DECODER_WRITE_CONTINUE;
+	return FLAC__STREAM_DECODER_WRITE_STATUS_CONTINUE;
 }
 
-void verify_metadata_callback(const FLAC__StreamDecoder *decoder, const FLAC__StreamMetaData *metadata, void *client_data)
+void verify_metadata_callback(const FLAC__StreamDecoder *decoder, const FLAC__StreamMetadata *metadata, void *client_data)
 {
 	(void)decoder;
 	(void)metadata;
